/*
 * graph.c
 * Copyright (C) 1994 by candy
 */
static char rcsid[] = "$Id: graph.c,v 1.2 1994/10/26 04:26:47 candy Exp candy $";
#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

#include "graph.h"

int graph_init_done;
Widget graph_widget;
Pixmap graph_pix;
GC graph_gc;
GC graph_bg;
unsigned int graph_width;
unsigned int graph_height;

int graph_npalet;
unsigned long *graph_pixels; /* graph_pixels[graph_npalet] */
XColor *graph_colors;

void
graph_get_graph_size(Widget w, long *xp, long *yp)
{
	Display *d = XtDisplay(w);
	Window win = XtWindow(w);
	Window root;
	int x, y;
	unsigned int width, height, border, depth;
	XGetGeometry(d, win, &root, &x, &y, &width, &height, &border, &depth);
	*xp = width;
	*yp = height;
}/* graph_get_graph_size */

void
graph_line(Widget w, int x0, int y0, int x1, int y1, int pal)
{
	Display *d = XtDisplay(w);
	if (pal == 0) {
		XDrawLine(d, graph_pix, graph_bg, x0, y0, x1, y1);
		XDrawLine(d, XtWindow(w), graph_bg, x0, y0, x1, y1);
	}
	else {
		XSetForeground(d, graph_gc, graph_pixels[pal]);
		XDrawLine(d, graph_pix, graph_gc, x0, y0, x1, y1);
		XDrawLine(d, XtWindow(w), graph_gc, x0, y0, x1, y1);
	}
}/* graph_line */

void
graph_box_fill(Widget w, int x0, int y0, int x1, int y1, int pal)
{
	Display *d = XtDisplay(w);
	int dx = x1 - x0, dy = y1 - y0;
	if (pal == 0) {
		XFillRectangle(d, graph_pix, graph_bg, x0, y0, dx, dy);
		XFillRectangle(d, XtWindow(w), graph_bg, x0, y0, dx, dy);
	}
	else {
		XSetForeground(d, graph_gc, graph_pixels[pal]);
		XFillRectangle(d, graph_pix, graph_gc, x0, y0, dx, dy);
		XFillRectangle(d, XtWindow(w), graph_gc, x0, y0, dx, dy);
	}
}/* graph_box_fill */

void
graph_circle(Widget w, int x0, int y0, int r, int pal)
{
	Display *d = XtDisplay(w);
	XSetForeground(d, graph_gc, graph_pixels[pal]);
	XDrawArc(d, graph_pix, graph_gc, x0 - r, y0 - r, 2 * r, 2 * r, 0, 360 * 64);
	XDrawArc(d, XtWindow(w), graph_gc, x0 - r, y0 - r, 2 * r, 2 * r, 0, 360 * 64);
}/* graph_circle */

void
graph_clear(Widget w)
{
	long wx, wy;
	graph_get_graph_size(w, &wx, &wy);
	graph_box_fill(w, 0, 0, wx, wy, 0);
}/* graph_clear */

void
graph_set_palet(Widget w, XColor *cp)
{
	if (graph_colors != NULL) {
		Display *d = XtDisplay(w);
		int scno = XScreenNumberOfScreen(XtScreen(w));
		XStoreColors(d, DefaultColormap(d, scno), cp, graph_npalet);
		if (cp != graph_colors) {
			int i;
			for (i = 0; i < graph_npalet; i++) {
				graph_colors[i].red = cp[i].red;
				graph_colors[i].green = cp[i].green;
				graph_colors[i].blue = cp[i].blue;
			}/* for */
		}
	}
}/* graph_set_palet */

void
graph_rotate_palet(Widget w, int start, unsigned int n, int dir, int idx, int red, int green, int blue)
{
	if (graph_colors != NULL) {
		if (start > graph_npalet) {
			/* ignore */
		}
		else {
			if (start + n > graph_npalet) {
				n = graph_npalet - start;
			}
			if (dir > 0) {
				int i;
				XColor t = graph_colors[start + n - 1];
				for (i = start + n - 1; i != start; i--) {
					graph_colors[i].red = graph_colors[i - 1].red;
					graph_colors[i].green = graph_colors[i - 1].green;
					graph_colors[i].blue = graph_colors[i - 1].blue;
				}/* for */
				graph_colors[i].red = t.red;
				graph_colors[i].green = t.green;
				graph_colors[i].blue = t.blue;
			}
			else {
				int i;
				XColor t = graph_colors[start];
				for (i = start; i < start + n - 1; i++) {
					graph_colors[i].red = graph_colors[i + 1].red;
					graph_colors[i].green = graph_colors[i + 1].green;
					graph_colors[i].blue = graph_colors[i + 1].blue;
				}/* for */
				graph_colors[i].red = t.red;
				graph_colors[i].green = t.green;
				graph_colors[i].blue = t.blue;
			}
			if (idx >= 0 && start + idx < graph_npalet) {
				int p = start + idx;
				graph_colors[p].red = red;
				graph_colors[p].green = green;
				graph_colors[p].blue = blue;
			}
			graph_set_palet(w, graph_colors);
		}
	}
}/* graph_rotate_palet */

void
graph_alloc_palet(Widget w, int npalet, unsigned long bg, unsigned long fg)
{
	Display *d = XtDisplay(w);
	int scno = XScreenNumberOfScreen(XtScreen(w));
	unsigned long plane_mask[1];
	if (graph_pixels != NULL) {
		free(graph_pixels);
	}
	graph_pixels = malloc(sizeof(*graph_pixels) * npalet);
	graph_colors = malloc(sizeof(*graph_colors) * npalet);
	graph_npalet = npalet;
	if (graph_pixels != NULL && graph_colors != NULL) {
		if (XAllocColorCells(d, DefaultColormap(d, scno), False, plane_mask, 0, graph_pixels, graph_npalet) == 0) {
			/* error */
			int i;
			graph_pixels[0] = bg;
			for (i = 1; i < npalet; i++) {
				graph_pixels[i] = fg;
			}/* for */
			free(graph_colors);
			graph_colors = NULL;
		}
		else {
			int i;
			for (i = 0; i < npalet; i++) {
				graph_colors[i].pixel = graph_pixels[i];
				graph_colors[i].flags = DoRed | DoGreen | DoBlue;
				graph_colors[i].green = 65535 * ((i & 4) / 4);
				graph_colors[i].red = 65535 * ((i & 2) / 2);
				graph_colors[i].blue = 65535 * (i & 1);
			}/* for */
			XStoreColors(d, DefaultColormap(d, scno), graph_colors, npalet);
		}
	}
}/* graph_alloc_palet */

static Pixmap
create_off_screen(Widget w)
{
	Display *d = XtDisplay(w);
	Window win = XtWindow(w);
	Pixmap pix;
	unsigned int width, height;
	{
		Window root;
		int x, y;
		unsigned int border, depth;
		XGetGeometry(d, win, &root, &x, &y, &width, &height, &border, &depth);
		pix = XCreatePixmap(d, win, width, height, depth);
	}
	{
		GC gc = XCreateGC(d, pix, 0, NULL);
		unsigned long bg;
		XtVaGetValues(w, XtNbackground, &bg, NULL);
		XSetForeground(d, gc, bg);
		XFillRectangle(d, pix, gc, 0, 0, width, height);
		XFreeGC(d, gc);
	}
	return pix;
}/* create_off_screen */

static unsigned int
get_window_size(Widget w, unsigned int *xp, unsigned int *yp)
{
	Display *d = XtDisplay(w);
	Window win = XtWindow(w);
	Window root;
	int x, y;
	unsigned int depth, border;
	XGetGeometry(d, win, &root, &x, &y, xp, yp, &border, &depth);
	return *yp;
}/* get_window_size */

void
graph_set_a_palet(Widget w, int pal, int red, int green, int blue)
{
	if (graph_colors != NULL) {
		Display *d = XtDisplay(w);
		int scno = XScreenNumberOfScreen(XtScreen(w));
		XColor *cp = &graph_colors[pal];
		cp->red = red;
		cp->green = green;
		cp->blue = blue;
		XStoreColor(d, DefaultColormap(d, scno), cp);
	}
}/* graph_set_a_palet */

void
graph_init_once(Widget w)
{
	if (!graph_init_done) {
		Display *d = XtDisplay(w);
		graph_init_done = 1;
		graph_widget = w;
		get_window_size(w, &graph_width, &graph_height);
		graph_pix = create_off_screen(w);
		graph_gc = XCreateGC(d, graph_pix, 0, NULL);
		graph_bg = XCreateGC(d, graph_pix, 0, NULL);
		{
			unsigned long fg, bg;
			XtVaGetValues(w, XtNborderColor, &fg, XtNbackground, &bg, NULL);
			if (fg == bg) {
				int scno = XScreenNumberOfScreen(XtScreen(w));
				if (fg == BlackPixel(d, scno))
					fg = WhitePixel(d, scno);
				else
					fg = BlackPixel(d, scno);
			}
			XSetForeground(d, graph_gc, fg);
			XSetForeground(d, graph_bg, bg);
			graph_alloc_palet(w, 16, bg, fg);
		}
	}
}/* graph_init_once */

#define MSG(s__)
#ifndef MSG
#define MSG(s__) printf s__
#endif

void
graph_create_proc(Widget w, XEvent *e, String *para, Cardinal *cp)
{
	MSG(("create_proc\n"));
	graph_init_once(w);
}/* graph_create_proc */

void
graph_map_proc(Widget w, XEvent *e, String *para, Cardinal *cp)
{
	MSG(("map_proc\n"));
	graph_init_once(w);
}/* graph_map_proc */

void
graph_config_proc(Widget w, XEvent *e, String *para, Cardinal *cp)
{
	unsigned int neww, newh;
	MSG(("config_proc\n"));
	graph_init_once(w);
	get_window_size(w, &neww, &newh);
	if (neww > graph_width || newh > graph_height) {
		Display *d = XtDisplay(w);
		Pixmap pix = create_off_screen(w);
		XCopyArea(d, graph_pix, pix, graph_gc, 0, 0, neww, newh, 0, 0);
		XFreePixmap(d, graph_pix);
		graph_pix = pix;
		graph_width = neww;
		graph_height = newh;
	}
}/* graph_config_proc */

void
graph_expose_proc(Widget w, XEvent *e, String *para, Cardinal *cp)
{
	Display *d = XtDisplay(w);
	Window win = XtWindow(w);
	int x0 = e->xexpose.x, y0 = e->xexpose.y, x1 = e->xexpose.width, y1 = e->xexpose.height;
	MSG(("expose_proc\n"));
	graph_init_once(w);
	XCopyArea(d, graph_pix, win, graph_gc, x0, y0, x1, y1, x0, y0);
}/* graph_expose_proc */

int
end_graph(void)
{
	return 0;
}/* end_graph */
